/*
 * Copyright (c) 2007 Massachusetts Institute of Technology
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.  
 *
 * * Neither the name of the Massachusetts Institute of Technology 
 *   nor the names of its contributors may be used to endorse or 
 *   promote products derived from this software without specific  
 *   prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package blog;

import java.util.*;

/**
 * A term consisting of a single logical variable.  VariableTerm objects 
 * are not created by the parser; instead, they're returned by the 
 * getTermInScope method on SymbolTerm.  
 */
public class VariableTerm extends Term {
    /**
     * Creates a new VariableTerm with the given logical variable.
     */
    public VariableTerm(LogicalVar var) {
	this.var = var;
    }

    public LogicalVar logicalVar() {
	return var;
    }

    public boolean checkTypesAndScope(Model model, Map scope) {
	if (var != scope.get(var.getName())) {
	    System.err.println(getLocation() + ": LogicalVar " + var 
			       + " is not in scope.");
	    return false;
	}
	return true;
    }
	
    public int compile(LinkedHashSet callStack) {
	return 0; // no compilation necessary for logical variables
    }

    public Type getType() {
	return var.getType();
    }

    public Object evaluate(EvalContext context) {
	return context.getLogicalVarValue(var);
    }

    public BayesNetVar getVariable() {
	return new DerivedVar(this);
    }

    public boolean containsRandomSymbol() {
	return false;
    }

    public Set getFreeVars() {
	return Collections.singleton(var);
    }

    public boolean equals(Object o) {
	if (o instanceof VariableTerm) {
	    VariableTerm other = (VariableTerm) o;
	    return (var == other.logicalVar());
	}
	return false;
    }

    public int hashCode() {
	return var.hashCode();
    }

    public String toString() {
	return var.getName();
    }

    private LogicalVar var;
}
